# frozen_string_literal: true

require 'weakref'

module Heimdall
  # Protocol error class
  class ProtoError < StandardError
  end

  # Container for any uniquely identifiable object
  # @abstract
  class UUIDObject
    UUIDSYM = (('0'..'9').to_a + ('a'..'f').to_a).freeze
    class << self
      # Add uuids to subclass
      def inherited(subclass)
        @uuids ||= {}
        @foreign_ids ||= {}
        @last ||= 0
        super(subclass)
        subclass.uuids = @uuids
        subclass.foreign_ids = @foreign_ids
      end

      # Create a new UUIDObject with unique UUID
      # @return [self]
      def new(*args, **params)
        object = super(*args, **params)
        @uuids[object.uuid] = WeakRef.new(object)
        @last = object.uuid
        add_foreign(object.foreign_ids, object)
        object
      end

      # Delete an object found by UUID
      # @param uuid [Integer]
      # @return [void]
      def delete(uuid)
        obj = @uuids[uuid]
        return false unless obj

        @uuids.delete(obj.uuid)
        obj.foreign_ids.each { |x| @foreign_ids.delete(x) }
        true
      end

      # Delete an object found by foreign id
      # @param foreign_id [String]
      # @return [void]
      def delete_foreign(foreign_id)
        obj = @foreign_ids[foreign_id]
        return false unless obj

        @uuids.delete(obj.uuid)
        obj.foreign_ids.each { |x| @foreign_ids.delete(x) }
        true
      end

      # Get object by UUID
      # @param uuid [Integer]
      # @return [self, nil]
      def get(uuid)
        @uuids[uuid] if @uuids[uuid].is_a? self
      end

      # Get object by foreign id
      # @param foreign_id [String]
      # @return [self, nil]
      def get_foreign(foreign_id)
        @foreign_ids[foreign_id] if @foreign_ids[foreign_id].is_a? self
      end

      # Add foreign ids to the list of known foreign ids
      # @param foreign [Array<String>]
      # @return [void]
      def add_foreign(foreign, obj)
        foreign.each do |f_id|
          @foreign_ids[check_foreign(f_id)] = WeakRef.new(obj)
        end
      end

      # Check a foreign id for valid syntax
      # @param foreign [String]
      # @return [String]
      def check_foreign(foreign)
        if foreign_ids.include?(foreign)
          raise ProtoError, 'foreign id already exists'
        end

        unless foreign.match?(/^([\w_]+):.*$/)
          raise ProtoError, 'invalid foreign id syntax'
        end

        foreign
      end

      attr_accessor :uuids, :foreign_ids, :last
    end

    def initialize
      @uuid = __gen_uuid
      @foreign_ids = []
      @foreign_ids.append(@id) if @id
    end

    attr_reader :uuid, :foreign_ids

    private

    def __gen_uuid
      newuuid = (Time.now.to_f * 1000).to_i * 10000
      if self.class.last && (self.class.last / 1000) == (newuuid / 1000)
        newuuid += (self.class.last % 1000) + 1
      end
      newuuid
    end
  end

  # PubSub abstract container
  # @abstract
  class PubSub
    def initialize(buffer_size: 1024)
      @buffer_size = buffer_size
      @listeners = []
      @messages = []
    end

    # Push message to all listeners
    # @param msg [UUIDObject]
    # @return [void]
    def push(msg)
      @listeners.each { |x| x.call(msg) }
      @messages.append(msg)
      @messages.shift if @messages.length > @buffer_size
    end

    # Add a listener to the PubSub
    # @param listener [#call]
    # @return [void]
    def listen(&listener)
      @listeners.append(listener)
    end

    # Pull all messages since UUID
    # @param uuid [String]
    # @return [Array<UUIDObject>]
    def pull(uuid)
      output = []
      @messages.reverse_each do |x|
        break if x.uuid < uuid

        output.unshift(x)
      end
      output
    end
  end

  # Message struct
  class Message < UUIDObject
    def initialize(datahash, **params)
      @id = datahash["id"]
      @from = if datahash["from"]
                UUIDObject.get(datahash["from"])
              elsif datahash["from_foreign"]
                UUIDObject.get_foreign(datahash["from_foreign"])
              end
      @to = if datahash["to"]
              UUIDObject.get(datahash["to"])
            elsif datahash["to_foreign"]
              UUIDObject.get_foreign(datahash["to_foreign"])
            end
      @content = datahash["content"]
      # @reply_to = datahash["reply_to"] # TODO: make this make sense
      @attachments = datahash["attachments"]

      super(**params)
    end

    # Convert message data to a JSON struct
    # @return [String] JSON struct
    def to_struct
      JSON.dump({
                  "id" => @id,
                  "uuid" => @uuid,
                  "from" => @from.uuid,
                  "to" => @to.uuid,
                  "from_foreign" => @from.foreing_ids,
                  "to_foreign" => @to.foreign_ids,
                  "content" => @content,
                  "attachments" => @attachments
                })
    end

    attr_reader :from, :to, :content, :reply_to, :attachments
  end

  # User struct
  class User < UUIDObject
    DEFAULT_AVATAR = ""

    def initialize(datahash, **params)
      @id = datahash["id"]
      @username = datahash["username"]
      @nickname = datahash["nickname"]
      @avatar = datahash["avatar"] || self.class::DEFAULT_AVATAR

      super(**params)
    end

    # Convert user data to a JSON struct
    # @return [String] JSON struct
    def to_struct
      JSON.dump({
                  "id" => @id,
                  "uuid" => @uuid,
                  "username" => @username,
                  "nickname" => @nickname,
                  "avatar" => @avatar
                })
    end
  end

  # Channel struct
  class Channel < UUIDObject
    DEFAULT_AVATAR = ""

    def initialize(datahash, **params)
      @pubsub = PubSub.new
      @id = datahash["id"]
      @name = datahash["name"] or @id
      @avatar = datahash["avatar"] || self.class::DEFAULT_AVATAR

      super(**params)
    end

    # Convert channel data to a JSON struct
    # @return [String] JSON struct
    def to_struct
      JSON.dump({
                  "id" => @id,
                  "uuid" => @uuid,
                  "name" => @name,
                  "avatar" => @avatar
                })
    end

    attr_reader :pubsub
  end

  VERSION = "1.0"
end
