# frozen_string_literal: true

require_relative '.env'
require_relative 'proto'
require 'landline'
require 'landline/extensions/websocket'
require 'json'

# Class for handling websocket connection requests
class Reactor
  def initialize
    @sockets = []
    @events = {}
  end

  # Attach websocket to reactor
  # @param socket [Landline::WebSocket::WSockWrapper]
  def attach(socket)
    @sockets.append(socket)
  end

  # Push data to all attached sockets
  # @param data [Array,Hash]
  def push(data)
    @sockets.each { |x| x.write(data) }
  end

  # Handle websocket event
  # @param event [String]
  # @param block [#call]
  def on(event, &block)
    @events[event] ||= []
    @events[event].append(block)
  end

  # Main loop
  def listen_loop
    loop do
      readable = IO.select(@sockets, nil, nil, 5)&.fetch(0, nil)
      readable&.each do |socket|
        output = socket.read
        if output
          _respond(socket, output)
        else
          @sockets.delete(socket)
        end
      end
    end
  end

  private

  def _respond(socket, output)
    data = JSON.parse(output.data)
    event_name = data['event']
    @events[event_name]&.each do |callback|
      callback.call(socket,
                    *(data['args'] || []),
                    **(data['params'] || {}))
    end
  rescue JSON::ParserError => e
    socket.write(JSON.dump({ error: e.message, code: 400 }))
  end
end

# Class acting as session-persistent object storage
class ObjectStorage
  def initialize
    @channels = {}
    @users = {}
  end

  # Add channel to object storage
  # @param channel [Heimdall::Channel]
  def add_channel(channel)
    @channels[channel.uuid] = channel
  end

  # Add user to object storage
  # @param user [Heimdall::User]
  def add_user(user)
    @users[user.uuid] = user
  end

  attr_reader :channels, :users
end

OBJECT_STORAGE = ObjectStorage.new

# Helper methods for building reactor listeners
module ReactorHelpers
  # Create a block for foreign and uuid indexing of a class
  # @param cls [Class]
  # @param block [#call]
  # @param reactor [Reactor]
  def self.foreign_and_uuid(method_name, reactor, cls, &block)
    reactor.on "#{cls.name.downcase}.#{method_name}.foreign" do |socket, f_id,
                                                                 *args, **vars|
      obj = cls.get_foreign(f_id)
      block.call(socket, obj, *args, **vars)
    end
    reactor.on "#{cls.name.downcase}.#{method_name}.uuid" do |socket, uuid,
                                                              *args, **vars|
      obj = cls.get_uuid(uuid)
      block.call(socket, obj, *args, **vars)
    end
  end

  # Boilerplate get/delete functions for a class
  # @param reactor [Reactor]
  # @param cls [Class]
  def self.foreign_and_uuid_boilerplate(reactor, cls)
    cname = cls.name.gsub(/\A.+::/, "").downcase
    reactor.on "#{cname}.get.foreign" do |socket, f_id|
      obj = cls.get_foreign(f_id)
      socket.write(obj || JSON.dump({ error: "#{cname} not found", code: 400 }))
    end
    reactor.on "#{cname}.get.uuid" do |socket, uuid|
      obj = cls.get_uuid(uuid)
      socket.write(obj || JSON.dump({ error: "#{cname} not found", code: 400 }))
    end
    reactor.on "#{cname}.delete.foreign" do |socket, f_id|
      status = cls.delete_foreign(f_id)
      socket.write(JSON.dump(if status
                               { code: 200 }
                             else
                               { error: "#{cname} not found", code: 400 }
                             end))
    end
    reactor.on "#{cname}.delete.uuid" do |socket, f_id|
      status = cls.delete_uuid(f_id)
      socket.write(JSON.dump(if status
                               { code: 200 }
                             else
                               { error: "#{cname} not found", code: 400 }
                             end))
    end
  end
end

# WebSocket handling
WEBSOCKET_LISTENER = Reactor.new.tap do |react|
  react.on "version" do |socket|
    socket.write(JSON.dump({
                             "version" => Heimdall::VERSION
                           }))
  end
  # --- USER ---
  ReactorHelpers.foreign_and_uuid_boilerplate(react, Heimdall::User)
  # --- CHANNEL ---
  ReactorHelpers.foreign_and_uuid_boilerplate(react, Heimdall::Channel)
  ReactorHelpers.foreign_and_uuid("listen", react,
                                  Heimdall::Channel) do |socket, channel|
    channel.pubsub.listen do |msg|
      socket.write(msg)
    end
    socket.write(JSON.dump({ code: 200 }))
  end
  ReactorHelpers.foreign_and_uuid("send", react,
                                  Heimdall::Channel) do |socket, channel,
                                                         **message_args|
    msg = Heimdall::Message.new(message_args)
    channel.pubsub.push(msg.to_struct)
    socket.write(JSON.dump({ code: 200,
                             msg_uuid: msg.uuid }))
  end
end

WEBSOCKET_LISTENER_LOOP = Thread.new { WEBSOCKET_LISTENER.listen_loop }

# Primary server class
class HeimdallServer < Landline::App
  before do
    header "content-type", "application/json"

    # Match data type against a list of datatypes
    # @param obj [Object]
    # @param type [Array, Class]
    def match_type(obj, type)
      if type.is_a? Array
        type.any? { |t| obj.is_a? t }
      else
        obj.is_a? type
      end
    end

    # Validate json body for a post request
    # @param args [Hash] hash of key - type pairs to check JSON data against
    def validate_json(**args)
      die(400, backtrace: ['JSON body expected']) unless json?
      @data = begin
        JSON.parse(request.body)
      rescue StandardError
        die(400, backtrace: ['JSON body is invalid'])
      end
      args.each do |k, v|
        unless match_type(@data[k], v)
          die(400, backtrace: ["Key #{k} of type #{v} is missing"])
        end
      end
    end
  end

  filter do
    defined? ::BOT_TOKEN ? request.cookies["token"] == ::BOT_TOKEN : true
  end

  get "/version" do
    JSON.dump({
                "version" => Heimdall::VERSION
              })
  end

  root __dir__
  serve "*.html"

  get "/debug" do
    jump "/debug.html"
  end

  pipeline do |request, &output|
    output.call(request)
  rescue Heimdall::ProtoError => e
    throw :finish, [400,
                    { "content-type": "application/json" },
                    JSON.dump({
                                "error" => e.message,
                                "code" => 400
                              })]
  end

  read_delete = proc do |cls|
    get "/foreign/*" do |id|
      user = cls.get_foreign(id)
      user&.to_struct or die(400, backtrace: ["#{cls} not found"])
    end

    get "/uuid/*" do |id|
      user = cls.get(id.to_i)
      user&.to_struct or die(400, backtrace: ["#{cls} not found"])
    end

    delete "/foreign/*" do |id|
      if cls.delete_foreign(id)
        JSON.dump({ "code" => 200 })
      else
        die(400, backtrace: ["#{cls} not found"])
      end
    end

    delete "/uuid/*" do |id|
      if cls.delete(id.to_i)
        JSON.dump({ "code" => 200 })
      else
        die(400, backtrace: ["#{cls} not found"])
      end
    end
  end

  path "/user" do
    post "/register" do
      validate_json("id" => String,
                    "username" => String,
                    "nickname" => [String, NilClass],
                    "avatar" => [String, NilClass])
      new_user = Heimdall::User.new(@data)
      OBJECT_STORAGE.add_user(new_user)
      JSON.dump({ "uuid": new_user.uuid })
    end

    instance_exec(Heimdall::User, &read_delete)
  end

  path "/channel" do
    post "/register" do
      validate_json("id" => String,
                    "name" => String,
                    "avatar" => [String, NilClass])
      new_channel = Heimdall::Channel.new(@data)
      OBJECT_STORAGE.add_channel(new_channel)
      JSON.dump({ "uuid": new_channel.uuid })
    end

    instance_exec(Heimdall::Channel, &read_delete)
  end

  websocket "/socket" do |socket|
    WEBSOCKET_LISTENER.attach(socket)
  end

  handle do |status, backtrace: nil, **_|
    backtrace ||= [Landline::Util::HTTP_STATUS[status]]
    [status,
     {
       "content-type": "application/json"
     },
     JSON.dump({
                 "error" => (backtrace || []).join("\n"),
                 "code" => status
               })]
  end
end

run HeimdallServer.new
